import concurrent.futures
import os
import yaml

from semantic_version import Version

from catalog_validation.schema.variable import Variable
from .exceptions import CatalogDoesNotExist, ValidationErrors
from .utils import validate_key_value_types, VALID_TRAIN_REGEX, WANTED_FILES_IN_ITEM_VERSION


def validate_catalog(catalog_path):
    if not os.path.exists(catalog_path):
        raise CatalogDoesNotExist(catalog_path)

    verrors = ValidationErrors()
    items = []
    item_futures = []
    for file_dir in os.listdir(catalog_path):
        complete_path = os.path.join(catalog_path, file_dir)
        if file_dir.startswith('.') or not os.path.isdir(complete_path) or file_dir in ('library', 'docs'):
            continue
        try:
            validate_train_structure(complete_path)
        except ValidationErrors as e:
            verrors.extend(e)
        else:
            items.extend(get_train_items(complete_path))

    with concurrent.futures.ProcessPoolExecutor(max_workers=5 if len(items) > 10 else 2) as exc:
        for item in items:
            item_futures.append(exc.submit(validate_catalog_item, item[0], item[1]))

        for future in item_futures:
            try:
                future.result()
            except ValidationErrors as e:
                verrors.extend(e)

    verrors.check()


def validate_train_structure(train_path):
    train = os.path.basename(train_path)
    verrors = ValidationErrors()
    if not VALID_TRAIN_REGEX.match(train):
        verrors.add(train, 'Train name is invalid.')

    verrors.check()


def get_train_items(train_path):
    train = os.path.basename(train_path)
    items = []
    for catalog_item in os.listdir(train_path):
        item_path = os.path.join(train_path, catalog_item)
        if not os.path.isdir(item_path):
            continue
        items.append((item_path, f'{train}.{catalog_item}'))
    return items


def validate_catalog_item(catalog_item_path, schema, validate_versions=True):
    # We should ensure that each catalog item has at least 1 version available
    # Also that we have item.yaml present
    verrors = ValidationErrors()
    item_name = os.path.join(catalog_item_path)
    files = []
    versions = []

    if not os.path.isdir(catalog_item_path):
        verrors.add(schema, 'Catalog item must be a directory')
    verrors.check()

    for file_dir in os.listdir(catalog_item_path):
        complete_path = os.path.join(catalog_item_path, file_dir)
        if os.path.isdir(complete_path):
            versions.append(complete_path)
        else:
            files.append(file_dir)

    if not versions:
        verrors.add(f'{schema}.versions', f'No versions found for {item_name} item.')

    if 'item.yaml' not in files:
        verrors.add(f'{schema}.item', 'Item configuration (item.yaml) not found')
    else:
        with open(os.path.join(catalog_item_path, 'item.yaml'), 'r') as f:
            item_config = yaml.safe_load(f.read())

        validate_key_value_types(item_config, (('categories', list),), verrors, f'{schema}.item_config')

    for version_path in (versions if validate_versions else []):
        try:
            validate_catalog_item_version(version_path, f'{schema}.versions.{os.path.basename(version_path)}')
        except ValidationErrors as e:
            verrors.extend(e)

    verrors.check()


def validate_catalog_item_version(version_path, schema):
    verrors = ValidationErrors()
    version_name = os.path.basename(version_path)
    item_name = version_path.split('/')[-2]
    try:
        Version(version_name)
    except ValueError:
        verrors.add(f'{schema}.name', f'{version_name!r} is not a valid version name.')

    files_diff = WANTED_FILES_IN_ITEM_VERSION ^ set(
        f for f in os.listdir(version_path) if f in WANTED_FILES_IN_ITEM_VERSION
    )
    if files_diff:
        verrors.add(f'{schema}.required_files', f'Missing {", ".join(files_diff)} required configuration files.')

    chart_version_path = os.path.join(version_path, 'Chart.yaml')
    if os.path.exists(chart_version_path):
        with open(chart_version_path, 'r') as f:
            try:
                chart_config = yaml.safe_load(f.read())
            except yaml.YAMLError:
                verrors.add(schema, 'Must be a valid yaml file')
            else:
                if not isinstance(chart_config, dict):
                    verrors.add(schema, 'Must be a dictionary')
                else:
                    if chart_config.get('name') != item_name:
                        verrors.add(f'{schema}.item_name', 'Item name not correctly set in "Chart.yaml".')

                    if chart_config.get('version') != version_name:
                        verrors.add(
                            f'{schema}.version',
                            'Configured version in "Chart.yaml" does not match version directory name.'
                        )

    questions_path = os.path.join(version_path, 'questions.yaml')
    if os.path.exists(questions_path):
        try:
            validate_questions_yaml(questions_path, f'{schema}.questions_configuration')
        except ValidationErrors as v:
            verrors.extend(v)

    verrors.check()


def validate_questions_yaml(questions_yaml_path, schema):
    verrors = ValidationErrors()

    with open(questions_yaml_path, 'r') as f:
        try:
            questions_config = yaml.safe_load(f.read())
        except yaml.YAMLError:
            verrors.add(schema, 'Must be a valid yaml file')
        else:
            if not isinstance(questions_config, dict):
                verrors.add(schema, 'Must be a dictionary')

    verrors.check()

    validate_key_value_types(
        questions_config, (('groups', list), ('questions', list), ('portals', dict, False)), verrors, schema
    )

    groups = []
    if type(questions_config.get('groups')) == list:
        for index, group in enumerate(questions_config['groups']):
            if type(group) != dict:
                verrors.add(f'{schema}.groups.{index}', 'Type of group should be a dictionary.')
                continue

            if group.get('name'):
                groups.append(group['name'])

            validate_key_value_types(group, (('name', str), ('description', str)), verrors, f'{schema}.group.{index}')

    if type(questions_config.get('questions')) != list:
        # We only want to raise verrors here if questions is not a list as otherwise we can raise them at the end
        # after validating the questions
        verrors.check()

    if isinstance(questions_config.get('portals'), dict):
        for index, portal_details in enumerate(questions_config['portals'].items()):
            portal_type, portal_schema = portal_details
            error_schema = f'{schema}.portals.{index}'
            if not isinstance(portal_type, str):
                verrors.add(error_schema, 'Portal type must be a string')
            if not isinstance(portal_schema, dict):
                verrors.add(error_schema, 'Portal schema must be a dictionary')
            else:
                validate_key_value_types(
                    portal_schema, (('protocols', list), ('host', list), ('ports', list), ('path', str, False)),
                    verrors, error_schema
                )

    for index, question in enumerate(questions_config['questions']):
        validate_question(question, f'{schema}.questions.{index}', verrors, (('group', str),))
        if question.get('group') and question['group'] not in groups:
            verrors.add(f'{schema}.questions.{index}.group', f'Please specify a group declared in "{schema}.groups"')

    verrors.check()


def validate_question(question_data, schema, verrors, validate_top_level_attrs=None):
    if type(question_data) != dict:
        verrors.add(schema, 'Question must be a valid dictionary.')
        return

    validate_top_level_attrs = validate_top_level_attrs or tuple()
    validate_key_value_types(
        question_data, (('variable', str), ('label', str), ('schema', dict)) + validate_top_level_attrs, verrors, schema
    )
    if type(question_data.get('schema')) != dict:
        return

    try:
        Variable(question_data).validate(schema)
    except ValidationErrors as ve:
        verrors.extend(ve)
        return

    schema_data = question_data['schema']
    variable_type = schema_data['type']

    for condition, key, schema_str in (
        (variable_type != 'list', 'subquestions', f'{schema}.schema.subquestions'),
        (variable_type == 'list', 'items', f'{schema}.schema.items'),
        (variable_type == 'dict', 'attrs', f'{schema}.schema.attrs'),
    ):
        if not (condition and type(schema_data.get(key)) == list):
            continue

        for index, item in enumerate(schema_data[key]):
            validate_question(item, f'{schema_str}.{index}', verrors)
